@@ -4,6 +4,7 @@ from __future__ import absolute_import, division, print_function
 
 import os
 import sys
+import weakref
 import warnings
 import webbrowser
 
@@ -11,7 +12,7 @@ from qtpy import QtCore, QtWidgets, QtGui, compat
 from qtpy.QtCore import Qt
 
 from glue.core.application_base import Application
-from glue.core.message import ApplicationClosedMessage
+from glue.core.message import ApplicationClosedMessage, DataCollectionMessage
 from glue.core import command, Data
 from glue.core.coordinates import WCSCoordinates
 from glue import env
@@ -36,7 +37,9 @@ from glue.app.qt.feedback import submit_bug_report, submit_feedback
 from glue.app.qt.plugin_manager import QtPluginManager
 from glue.app.qt.versions import QVersionsDialog
 from glue.app.qt.terminal import glue_terminal, IPythonTerminalError
-from glue.config import qt_fixed_layout_tab, qt_client, startup_action
+
+from glue.config import qt_fixed_layout_tab, qt_client, startup_action, keyboard_shortcut
+from glue.app.qt.save_data import SaveDataDialog
 
 __all__ = ['GlueApplication']
 DOCS_URL = 'http://www.glueviz.org'
@@ -100,10 +103,10 @@ class GlueLogger(QtWidgets.QWidget):
         self._text.setTextInteractionFlags(Qt.TextSelectableByMouse)
 
         clear = QtWidgets.QPushButton("Clear")
-        clear.clicked.connect(nonpartial(self._clear))
+        clear.clicked.connect(self._clear)
 
         report = QtWidgets.QPushButton("Send Bug Report")
-        report.clicked.connect(nonpartial(self._send_report))
+        report.clicked.connect(self._send_report)
 
         if isinstance(sys.stderr, GlueLogger):
             if isinstance(sys.stderr._stderr_original, GlueLogger):
@@ -149,14 +152,14 @@ class GlueLogger(QtWidgets.QWidget):
         """
         pass
 
-    def _send_report(self):
+    def _send_report(self, *args):
         """
         Send the contents of the log as a bug report
         """
         text = self._text.document().toPlainText()
         submit_bug_report(text)
 
-    def _clear(self):
+    def _clear(self, *args):
         """
         Erase the log
         """
@@ -183,6 +186,39 @@ class GlueLogger(QtWidgets.QWidget):
             sys.stderr = self._stderr_original
 
 
+class ExportHelper(object):
+    """
+    This class is needed because setting up the callbacks requires using
+    nonpartial but if the callback was a method on GlueApplication this would
+    result in a circular reference - hence we use a helper object with a
+    weak reference to the application.
+    """
+
+    def __init__(self, app):
+        self.app = weakref.ref(app)
+
+    @messagebox_on_error("Failed to export session")
+    def _choose_export_session(self, saver, checker, outmode):
+        app = self.app()
+        if app is None:
+            return
+        checker(app)
+        if outmode is None:
+            return saver(app)
+        elif outmode in ['file', 'directory']:
+            outfile, file_filter = compat.getsavefilename(parent=app)
+            if not outfile:
+                return
+            return saver(app, outfile)
+        else:
+            assert outmode == 'label'
+            label, ok = QtWidgets.QInputDialog.getText(app, 'Choose a label:',
+                                                       'Choose a label:')
+            if not ok:
+                return
+            return saver(app, label)
+
+
 class GlueApplication(Application, QtWidgets.QMainWindow):
 
     """ The main GUI application for the Qt frontend"""
@@ -197,6 +233,8 @@ class GlueApplication(Application, QtWidgets.QMainWindow):
         if self._original_app is not None:
             self._original_icon = self._original_app.windowIcon()
 
+        self._export_helper = ExportHelper(self)
+
         # Now we can get the application instance, which involves setting it
         # up if it doesn't already exist.
         self.app = get_qapp()
@@ -205,6 +243,9 @@ class GlueApplication(Application, QtWidgets.QMainWindow):
         Application.__init__(self, data_collection=data_collection,
                              session=session)
 
+        # Pull in any keybindings from an external file
+        self.keybindings = keyboard_shortcut
+
         icon = get_icon('app_icon')
         self.app.setWindowIcon(icon)
 
@@ -229,14 +270,47 @@ class GlueApplication(Application, QtWidgets.QMainWindow):
         lwidget = self._layer_widget
         a = PlotAction(lwidget, self)
         lwidget.ui.layerTree.addAction(a)
-        lwidget.bind_selection_to_edit_subset()
 
         self._tweak_geometry()
         self._create_actions()
         self._create_menu()
         self._connect()
         self.new_tab()
-        self._update_plot_dashboard(None)
+        self._update_viewer_in_focus()
+
+    def _update_viewer_in_focus(self, *args):
+
+        if not hasattr(self, '_viewer_in_focus'):
+            self._viewer_in_focus = None
+
+        mdi_area = self.current_tab
+        active = mdi_area.activeSubWindow()
+
+        # Disable any active tool in the viewer that was previously in focus.
+        # Note that we want to do this even if active is None, which means that
+        # the user may have switched application.
+        if (self._viewer_in_focus is not None and
+                (active is None or active.widget() is not self._viewer_in_focus)):
+            try:
+                self._viewer_in_focus.toolbar.active_tool = None
+            except AttributeError:
+                pass  # not all viewers have toolbars
+
+        if active is None:
+            first_viewer = None
+            for win in mdi_area.subWindowList():
+                if self._viewer_in_focus is win.widget():
+                    break
+                elif isinstance(win.widget(), DataViewer):
+                    first_viewer = win.widget()
+            else:
+                self._viewer_in_focus = first_viewer
+                self._update_focus_decoration()
+                self._update_plot_dashboard()
+        else:
+            self._viewer_in_focus = active.widget()
+            self._update_focus_decoration()
+            self._update_plot_dashboard()
 
     def run_startup_action(self, name):
         if name in startup_action.members:
@@ -250,7 +324,7 @@ class GlueApplication(Application, QtWidgets.QMainWindow):
         self.setCentralWidget(self._ui)
         self._ui.tabWidget.setTabBar(GlueTabBar())
 
-        lw = LayerTreeWidget()
+        lw = LayerTreeWidget(session=self._session)
         lw.set_checkable(False)
         self._vb = QtWidgets.QVBoxLayout()
         self._vb.setContentsMargins(0, 0, 0, 0)
@@ -268,15 +342,24 @@ class GlueApplication(Application, QtWidgets.QMainWindow):
         self._button_open_data.setText("Open Data")
         self._button_open_data.setIcon(get_icon('glue_open'))
         self._button_open_data.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)
-        self._button_open_data.clicked.connect(nonpartial(self._choose_load_data))
+        self._button_open_data.clicked.connect(self._choose_load_data_wizard)
 
         self._data_toolbar.addWidget(self._button_open_data)
 
+        self._button_save_data = QtWidgets.QToolButton()
+        self._button_save_data.setText("Export Data/Subsets")
+        self._button_save_data.setIcon(get_icon('glue_filesave'))
+        self._button_save_data.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)
+        self._button_save_data.clicked.connect(self._choose_save_data)
+
+        self._data_toolbar.addWidget(self._button_save_data)
+
         self._button_link_data = QtWidgets.QToolButton()
         self._button_link_data.setText("Link Data")
         self._button_link_data.setIcon(get_icon('glue_link'))
         self._button_link_data.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)
         self._button_link_data.clicked.connect(self._set_up_links)
+        self._on_data_collection_change()
 
         self._data_toolbar.addWidget(self._button_link_data)
 
@@ -289,21 +372,29 @@ class GlueApplication(Application, QtWidgets.QMainWindow):
 
         self._data_toolbar.addWidget(self._button_ipython)
 
-        self._button_open_data = QtWidgets.QToolButton()
-        self._button_open_data.setText("Open Session")
-        self._button_open_data.setIcon(get_icon('glue_open'))
-        self._button_open_data.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)
-        self._button_open_data.clicked.connect(nonpartial(self._restore_session))
+        self._button_open_session = QtWidgets.QToolButton()
+        self._button_open_session.setText("Open Session")
+        self._button_open_session.setIcon(get_icon('glue_open'))
+        self._button_open_session.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)
+        self._button_open_session.clicked.connect(self._restore_session)
 
-        self._data_toolbar.addWidget(self._button_open_data)
+        self._data_toolbar.addWidget(self._button_open_session)
 
-        self._button_open_data = QtWidgets.QToolButton()
-        self._button_open_data.setText("Save Session")
-        self._button_open_data.setIcon(get_icon('glue_filesave'))
-        self._button_open_data.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)
-        self._button_open_data.clicked.connect(nonpartial(self._choose_save_session))
+        self._button_save_session = QtWidgets.QToolButton()
+        self._button_save_session.setText("Export Session")
+        self._button_save_session.setIcon(get_icon('glue_filesave'))
+        self._button_save_session.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)
+        self._button_save_session.clicked.connect(self._choose_save_session)
 
-        self._data_toolbar.addWidget(self._button_open_data)
+        self._data_toolbar.addWidget(self._button_save_session)
+
+        self._button_edit_components = QtWidgets.QToolButton()
+        self._button_edit_components.setText("Add/edit arithmetic attributes")
+        self._button_edit_components.setIcon(get_icon('pencil'))
+        self._button_edit_components.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)
+        self._button_edit_components.clicked.connect(nonpartial(self._layer_widget._create_component))
+
+        self._data_toolbar.addWidget(self._button_edit_components)
 
         spacer = QtWidgets.QWidget()
         spacer.setMinimumSize(20, 10)
@@ -316,7 +407,7 @@ class GlueApplication(Application, QtWidgets.QMainWindow):
 
         # Selection mode toolbar
 
-        tbar = EditSubsetModeToolBar()
+        tbar = EditSubsetModeToolBar(parent=self)
         self._mode_toolbar = tbar
 
         self.addToolBar(self._mode_toolbar)
@@ -337,7 +428,7 @@ class GlueApplication(Application, QtWidgets.QMainWindow):
         self._button_preferences.setText("Preferences")
         self._button_preferences.setIcon(get_icon('glue_settings'))
         self._button_preferences.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)
-        self._button_preferences.clicked.connect(nonpartial(self._edit_settings))
+        self._button_preferences.clicked.connect(self._edit_settings)
 
         self._console_toolbar.addWidget(self._button_preferences)
 
@@ -354,6 +445,37 @@ class GlueApplication(Application, QtWidgets.QMainWindow):
         self._log.resize(550, 550)
         self._log.hide()
 
+        self._hub.subscribe(self, DataCollectionMessage, handler=self._on_data_collection_change)
+
+    def _on_data_collection_change(self, *event):
+        self._button_link_data.setEnabled(len(self.data_collection) > 1)
+
+    def keyPressEvent(self, event):
+        if self.current_tab.activeSubWindow() and self.current_tab.activeSubWindow().widget():
+            active_window = self.current_tab.activeSubWindow().widget()
+        else:
+            active_window = None
+
+        # keybindings is a data structure in the form of dict(dict)
+        # which uses the DataViewer as the first key, the key pressed
+        # as the second key, and the function associated with those two
+        # as the value.
+        if type(active_window) in self.keybindings:
+            for k, func in self.keybindings.members[type(active_window)].items():
+                if event.key() == k:
+                    func(self.session)
+                    return
+
+        # If key does not correspond with viewers, it might correspond
+        # with the global application, thus, None
+        if None in self.keybindings:
+            for k, func in self.keybindings.members[None].items():
+                if event.key() == k:
+                    func(self.session)
+                    return
+
+        return super(GlueApplication, self).keyPressEvent(event)
+
     def _set_up_links(self, event):
         LinkEditor.update_links(self.data_collection)
 
@@ -395,7 +517,7 @@ class GlueApplication(Application, QtWidgets.QMainWindow):
             return self.current_tab
         return self._ui.tabWidget.widget(index)
 
-    def new_tab(self):
+    def new_tab(self, *args):
         """Spawn a new tab page"""
         layout = QtWidgets.QGridLayout()
         layout.setSpacing(1)
@@ -406,7 +528,7 @@ class GlueApplication(Application, QtWidgets.QMainWindow):
         self._total_tab_count += 1
         tab.addTab(widget, str("Tab %i" % self._total_tab_count))
         tab.setCurrentWidget(widget)
-        widget.subWindowActivated.connect(self._update_plot_dashboard)
+        widget.subWindowActivated.connect(self._update_viewer_in_focus)
 
     def close_tab(self, index, warn=True):
         """ Close a tab window and all associated data viewers """
@@ -474,12 +596,13 @@ class GlueApplication(Application, QtWidgets.QMainWindow):
         pos = getattr(new_widget, 'position', None)
         sub = new_widget.mdi_wrap()
 
-        sub.closed.connect(self._clear_dashboard)
+        sub.closed.connect(self._update_viewer_in_focus)
 
         if label:
             sub.setWindowTitle(label)
         page.addSubWindow(sub)
         page.setActiveSubWindow(sub)
+
         if hold_position and pos is not None:
             new_widget.move(pos[0], pos[1])
 
@@ -487,17 +610,15 @@ class GlueApplication(Application, QtWidgets.QMainWindow):
 
         return sub
 
-    def _edit_settings(self):
+    def _edit_settings(self, *args):
         self._editor = PreferencesDialog(self, parent=self)
         self._editor.show()
 
-    def gather_current_tab(self):
+    def gather_current_tab(self, *args):
         """Arrange windows in current tab via tiling"""
         self.current_tab.tileSubWindows()
 
-    def _get_plot_dashboards(self, sub_window):
-
-        widget = sub_window.widget()
+    def _get_plot_dashboards(self, widget):
 
         if not isinstance(widget, DataViewer):
             return QtWidgets.QWidget(), QtWidgets.QWidget(), ""
@@ -522,14 +643,14 @@ class GlueApplication(Application, QtWidgets.QMainWindow):
         self._ui.plot_options_label.setText("Plot Options")
         self._ui.plot_layers_label.setText("Plot Layers")
 
-    def _update_plot_dashboard(self, sub_window):
+    def _update_plot_dashboard(self, *args):
+
         self._clear_dashboard()
 
-        if sub_window is None:
+        if self._viewer_in_focus is None:
             return
 
-        layer_view, options_widget, title = self._get_plot_dashboards(
-            sub_window)
+        layer_view, options_widget, title = self._get_plot_dashboards(self._viewer_in_focus)
 
         layout = self._ui.plot_layers.layout()
         layout.addWidget(layer_view)
@@ -551,18 +672,17 @@ class GlueApplication(Application, QtWidgets.QMainWindow):
 
     def _update_focus_decoration(self):
         mdi_area = self.current_tab
-        active = mdi_area.activeSubWindow()
-
         for win in mdi_area.subWindowList():
             widget = win.widget()
             if isinstance(widget, DataViewer):
-                widget.set_focus(win is active)
+                widget.set_focus(widget is self._viewer_in_focus)
 
     def _connect(self):
         self.setAcceptDrops(True)
         self._layer_widget.setup(self._data)
 
         self.tab_widget.tabCloseRequested.connect(self.close_tab)
+        self.tab_widget.currentChanged.connect(self._update_viewer_in_focus)
 
     def _create_menu(self):
         mbar = self.menuBar()
@@ -578,8 +698,9 @@ class GlueApplication(Application, QtWidgets.QMainWindow):
         menu.addAction(self._actions['session_reset'])
         menu.addAction(self._actions['session_restore'])
         menu.addAction(self._actions['session_save'])
+        menu.addAction(self._actions['export_data'])
         if 'session_export' in self._actions:
-            submenu = menu.addMenu("E&xport")
+            submenu = menu.addMenu("Advanced E&xporters")
             for a in self._actions['session_export']:
                 submenu.addAction(a)
         menu.addSeparator()
@@ -653,6 +774,9 @@ class GlueApplication(Application, QtWidgets.QMainWindow):
         window.show()
         window.exec_()
 
+    def _choose_load_data_wizard(self, *args):
+        self._choose_load_data(data_importer=data_wizard)
+
     def _choose_load_data(self, data_importer=None):
         if data_importer is None:
             self.add_datasets(self.data_collection, data_wizard())
@@ -667,6 +791,10 @@ class GlueApplication(Application, QtWidgets.QMainWindow):
                                     "Data objects")
             self.add_datasets(self.data_collection, data)
 
+    def _choose_save_data(self, *args):
+        dialog = SaveDataDialog(data_collection=self.data_collection, parent=self)
+        dialog.exec_()
+
     def _create_actions(self):
         """ Create and connect actions, store in _actions dict """
         self._actions = {}
@@ -674,7 +802,7 @@ class GlueApplication(Application, QtWidgets.QMainWindow):
         a = action("&New Data Viewer", self,
                    tip="Open a new visualization window in the current tab",
                    shortcut=QtGui.QKeySequence.New)
-        a.triggered.connect(nonpartial(self.choose_new_data_viewer))
+        a.triggered.connect(self._choose_new_data_viewer_nodata)
         self._actions['viewer_new'] = a
 
         if len(qt_client.members) == 0:
@@ -682,7 +810,7 @@ class GlueApplication(Application, QtWidgets.QMainWindow):
 
         a = action("New Fixed Layout Tab", self,
                    tip="Create a new tab with a fixed layout")
-        a.triggered.connect(nonpartial(self.choose_new_fixed_layout_tab))
+        a.triggered.connect(self.choose_new_fixed_layout_tab)
         self._actions['fixed_layout_tab_new'] = a
 
         if len(qt_fixed_layout_tab.members) == 0:
@@ -691,7 +819,7 @@ class GlueApplication(Application, QtWidgets.QMainWindow):
         a = action('New &Tab', self,
                    shortcut=QtGui.QKeySequence.AddTab,
                    tip='Add a new tab')
-        a.triggered.connect(nonpartial(self.new_tab))
+        a.triggered.connect(self.new_tab)
         self._actions['tab_new'] = a
 
         a = action('&Rename Tab', self,
@@ -703,20 +831,19 @@ class GlueApplication(Application, QtWidgets.QMainWindow):
         a = action('&Gather Windows', self,
                    tip='Gather plot windows side-by-side',
                    shortcut='Ctrl+G')
-        a.triggered.connect(nonpartial(self.gather_current_tab))
+        a.triggered.connect(self.gather_current_tab)
         self._actions['gather'] = a
 
-        a = action('&Save Session', self,
+        a = action('&Export Session', self,
                    tip='Save the current session')
-        a.triggered.connect(nonpartial(self._choose_save_session))
+        a.triggered.connect(self._choose_save_session)
         self._actions['session_save'] = a
 
         # Add file loader as first item in File menu for convenience. We then
         # also add it again below in the Import menu for consistency.
         a = action("&Open Data Set", self, tip="Open a new data set",
                    shortcut=QtGui.QKeySequence.Open)
-        a.triggered.connect(nonpartial(self._choose_load_data,
-                                       data_wizard))
+        a.triggered.connect(self._choose_load_data_wizard)
         self._actions['data_new'] = a
 
         # We now populate the "Import data" menu
@@ -726,15 +853,13 @@ class GlueApplication(Application, QtWidgets.QMainWindow):
 
         # Add default file loader (later we can add this to the registry)
         a = action("Import from file", self, tip="Import from file")
-        a.triggered.connect(nonpartial(self._choose_load_data,
-                                       data_wizard))
+        a.triggered.connect(self._choose_load_data_wizard)
         acts.append(a)
 
         for i in importer:
             label, data_importer = i
             a = action(label, self, tip=label)
-            a.triggered.connect(nonpartial(self._choose_load_data,
-                                           data_importer))
+            a.triggered.connect(self._choose_load_data_wizard)
             acts.append(a)
 
         self._actions['data_importers'] = acts
@@ -747,7 +872,7 @@ class GlueApplication(Application, QtWidgets.QMainWindow):
                 a = action(label, self,
                            tip='Export the current session to %s format' %
                            label)
-                a.triggered.connect(nonpartial(self._choose_export_session,
+                a.triggered.connect(nonpartial(self._export_helper._choose_export_session,
                                                saver, checker, mode))
                 acts.append(a)
 
@@ -755,25 +880,30 @@ class GlueApplication(Application, QtWidgets.QMainWindow):
 
         a = action('Open S&ession', self,
                    tip='Restore a saved session')
-        a.triggered.connect(nonpartial(self._restore_session))
+        a.triggered.connect(self._restore_session)
         self._actions['session_restore'] = a
 
         a = action('Reset S&ession', self,
                    tip='Reset session to clean state')
-        a.triggered.connect(nonpartial(self._reset_session))
+        a.triggered.connect(self._reset_session)
         self._actions['session_reset'] = a
 
+        a = action('Export D&ata/Subsets', self,
+                   tip='Export data to a file')
+        a.triggered.connect(self._choose_save_data)
+        self._actions['export_data'] = a
+
         a = action("Undo", self,
                    tip='Undo last action',
                    shortcut=QtGui.QKeySequence.Undo)
-        a.triggered.connect(nonpartial(self.undo))
+        a.triggered.connect(self.undo)
         a.setEnabled(False)
         self._actions['undo'] = a
 
         a = action("Redo", self,
                    tip='Redo last action',
                    shortcut=QtGui.QKeySequence.Redo)
-        a.triggered.connect(nonpartial(self.redo))
+        a.triggered.connect(self.redo)
         a.setEnabled(False)
         self._actions['redo'] = a
 
@@ -790,10 +920,16 @@ class GlueApplication(Application, QtWidgets.QMainWindow):
 
         a = action('&Plugin Manager', self,
                    tip='Open plugin manager')
-        a.triggered.connect(nonpartial(self.plugin_manager))
+        a.triggered.connect(self.plugin_manager)
         self._actions['plugin_manager'] = a
 
-    def choose_new_fixed_layout_tab(self):
+    def undo(self, *args):
+        super(GlueApplication, self).undo()
+
+    def redo(self, *args):
+        super(GlueApplication, self).redo()
+
+    def choose_new_fixed_layout_tab(self, *args):
         """
         Creates a new tab with a fixed layout
         """
@@ -815,10 +951,13 @@ class GlueApplication(Application, QtWidgets.QMainWindow):
             name += ': ' + tab.LABEL
         self.tab_widget.addTab(tab, name)
         self.tab_widget.setCurrentWidget(tab)
-        tab.subWindowActivated.connect(self._update_plot_dashboard)
+        tab.subWindowActivated.connect(self._update_viewer_in_focus)
 
         return tab
 
+    def _choose_new_data_viewer_nodata(self):
+        self.choose_new_data_viewer()
+
     def choose_new_data_viewer(self, data=None):
         """ Create a new visualization window in the current tab
         """
@@ -842,7 +981,7 @@ class GlueApplication(Application, QtWidgets.QMainWindow):
 
     new_data_viewer = defer_draw(Application.new_data_viewer)
 
-    def _choose_save_session(self):
+    def _choose_save_session(self, *args):
         """ Save the data collection and hub to file.
 
         Can be restored via restore_session
@@ -850,8 +989,11 @@ class GlueApplication(Application, QtWidgets.QMainWindow):
 
         # include file filter twice, so it shows up in Dialog
         outfile, file_filter = compat.getsavefilename(
-            parent=self, filters=("Glue Session (*.glu);; "
-                                  "Glue Session including data (*.glu)"))
+            parent=self, basedir=getattr(self, '_last_session_name', 'session.glu'),
+            filters=("Glue Session with absolute paths to data(*.glu);; "
+                     "Glue Session with relative paths to data(*.glu);; "
+                     "Glue Session including data (*.glu)"),
+            selectedfilter=getattr(self, '_last_session_filter', ''))
 
         # This indicates that the user cancelled
         if not outfile:
@@ -861,30 +1003,16 @@ class GlueApplication(Application, QtWidgets.QMainWindow):
         if '.' not in outfile:
             outfile += '.glu'
 
-        with set_cursor_cm(Qt.WaitCursor):
-            self.save_session(
-                outfile, include_data="including data" in file_filter)
+        self._last_session_name = outfile
+        self._last_session_filter = file_filter
 
-    @messagebox_on_error("Failed to export session")
-    def _choose_export_session(self, saver, checker, outmode):
-        checker(self)
-        if outmode is None:
-            return saver(self)
-        elif outmode in ['file', 'directory']:
-            outfile, file_filter = compat.getsavefilename(parent=self)
-            if not outfile:
-                return
-            return saver(self, outfile)
-        else:
-            assert outmode == 'label'
-            label, ok = QtWidgets.QInputDialog.getText(self, 'Choose a label:',
-                                                       'Choose a label:')
-            if not ok:
-                return
-            return saver(self, label)
+        with set_cursor_cm(Qt.WaitCursor):
+            self.save_session(outfile,
+                              include_data="including data" in file_filter,
+                              absolute_paths="absolute" in file_filter)
 
     @messagebox_on_error("Failed to restore session")
-    def _restore_session(self, show=True):
+    def _restore_session(self, *args):
         """ Load a previously-saved state, and restart the session """
         fltr = "Glue sessions (*.glu)"
         file_name, file_filter = compat.getopenfilename(
@@ -903,11 +1031,13 @@ class GlueApplication(Application, QtWidgets.QMainWindow):
         return (len([viewer for tab in self.viewers for viewer in tab]) == 0 and
                 len(self.data_collection) == 0)
 
-    def _reset_session(self, show=True, warn=True):
+    def _reset_session(self, *args, **kwargs):
         """
         Reset session to clean state.
         """
 
+        warn = kwargs.pop('warn', False)
+
         if not os.environ.get('GLUE_TESTING') and warn and not self.is_empty:
             buttons = QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel
             dialog = QtWidgets.QMessageBox.warning(
@@ -920,7 +1050,8 @@ class GlueApplication(Application, QtWidgets.QMainWindow):
 
         # Make sure the closeEvent gets executed to close the GlueLogger
         self._log.close()
-        self.app.processEvents()
+        if self.app is not None:
+            self.app.processEvents()
 
         ga = GlueApplication()
         ga.start(block=False)
@@ -1087,14 +1218,23 @@ class GlueApplication(Application, QtWidgets.QMainWindow):
 
     def closeEvent(self, event):
         """Emit a message to hub before closing."""
+
+        # Clear the namespace in the terminal to avoid cicular references
+        if self._terminal is not None:
+            self._terminal.widget().clear_ns(['data_collection', 'dc', 'hub', 'session', 'application'])
+
         for tab in self.viewers:
             for viewer in tab:
                 viewer.close(warn=False)
+        self._viewer_in_focus = None
+        self._clear_dashboard()
         self._log.close()
         self._hub.broadcast(ApplicationClosedMessage(None))
         event.accept()
         if self._original_app is not None:
             self._original_app.setWindowIcon(self._original_icon)
+            self._original_app = None
+        self.app = None
 
     def report_error(self, message, detail):
         """
@@ -1110,12 +1250,12 @@ class GlueApplication(Application, QtWidgets.QMainWindow):
         qmb.resize(400, qmb.size().height())
         qmb.exec_()
 
-    def plugin_manager(self):
+    def plugin_manager(self, *args):
         from glue.main import _installed_plugins
         pm = QtPluginManager(installed=_installed_plugins)
         pm.ui.exec_()
 
-    def _update_undo_redo_enabled(self):
+    def _update_undo_redo_enabled(self, *args):
         undo, redo = self._cmds.can_undo_redo()
         self._actions['undo'].setEnabled(undo)
         self._actions['redo'].setEnabled(redo)
