@@ -6,7 +6,6 @@ from abc import ABCMeta, abstractmethod
 
 from glue.utils import CallbackMixin
 from glue.core.data_factories import load_data
-from glue.core.edit_subset_mode import EditSubsetMode
 
 MAX_UNDO = 50
 """
@@ -286,8 +285,10 @@ class ApplySubsetState(Command):
         DataCollection to operate on
     subset_state: :class:`~glue.core.subset_state.SubsetState`
         Subset state to apply
+    use_current: bool
+        Flag indicating whether to update current subset or create a new one
     """
-    kwargs = ['data_collection', 'subset_state']
+    kwargs = ['data_collection', 'subset_state', 'use_current']
     label = 'apply subset'
 
     def do(self, session):
@@ -297,8 +298,8 @@ class ApplySubsetState(Command):
             for subset in data.subsets:
                 self.old_states[subset] = subset.subset_state
 
-        mode = EditSubsetMode()
-        mode.update(self.data_collection, self.subset_state)
+        mode = session.edit_subset_mode
+        mode.update(self.data_collection, self.subset_state, use_current=self.use_current)
 
     def undo(self, session):
         for data in self.data_collection:
