@@ -88,10 +88,10 @@ from glue.utils import lookup_class
 literals = tuple([type(None), float, int, bytes, bool])
 
 if six.PY2:
-    literals += (long,)
+    literals += (long,)  # noqa
 
 
-literals += np.ScalarType
+literals += tuple(s for s in np.ScalarType if s not in (np.datetime64, np.timedelta64))
 
 builtin_iterables = (tuple, list, set)
 
@@ -101,7 +101,7 @@ JSON_ENCODER = json.JSONEncoder()
 # classes/functions around in Glue, so we have a file that maps the old paths to
 # the new location, and we read this in to PATH_PATCHES.
 PATCH_FILE = os.path.abspath(os.path.join(os.path.dirname(__file__),
-                             'state_path_patches.txt'))
+                                          'state_path_patches.txt'))
 
 # For Mac app, need to get file from source directory
 if not os.path.exists(PATCH_FILE) and 'site-packages.zip' in PATCH_FILE:
@@ -113,6 +113,18 @@ for line in open(PATCH_FILE):
     PATH_PATCHES[before.strip()] = after.strip()
 
 
+def save(filename, obj):
+    s = GlueSerializer(obj)
+    with open(filename, 'w') as f:
+        s.dump(f)
+
+
+def load(filename):
+    with open(filename, 'r') as f:
+        s = GlueUnSerializer.load(f)
+    return s.object('__main__')
+
+
 def lookup_class_with_patches(name):
     """
     A wrapper to lookup_class that also patches paths to ensure
@@ -248,13 +260,14 @@ class GlueSerializer(object):
     """
     dispatch = VersionedDict()
 
-    def __init__(self, obj, include_data=False):
+    def __init__(self, obj, include_data=False, absolute_paths=True):
         self._names = {}  # map id(object) -> name
         self._objs = {}   # map name -> object
         self._working = set()
         self._main = obj
         self.id(obj)
         self.include_data = include_data
+        self.absolute_paths = absolute_paths
 
     @classmethod
     def serializes(cls, obj, version=1):
@@ -427,6 +440,7 @@ class GlueUnSerializer(object):
         self._objs = {}   # map name -> object
         self._working = set()
         self._rec = json.loads(string) if string else json.load(fobj)
+        self._callbacks = []
 
     @classmethod
     def loads(cls, string):
@@ -494,6 +508,9 @@ class GlueUnSerializer(object):
         func = self._dispatch(rec)
         obj = func(rec, self)
 
+        if hasattr(obj, '__setgluestate_callback__'):
+            self._callbacks.append(obj.__setgluestate_callback__)
+
         # loader functions might yield the constructed value,
         # and then futher populate it. This deals with circular
         # dependencies.
@@ -508,8 +525,19 @@ class GlueUnSerializer(object):
             for _ in gen:  # ... and finish constructing it
                 pass
 
+        self._try_callbacks()
+
         return obj
 
+    def _try_callbacks(self):
+        for callback in self._callbacks[:]:
+            try:
+                callback(self)
+            except Exception:
+                pass
+            else:
+                self._callbacks.remove(callback)
+
 
 saver = GlueSerializer.serializes
 loader = GlueUnSerializer.unserializes
@@ -518,7 +546,7 @@ loader = GlueUnSerializer.unserializes
 @saver(dict)
 def _save_dict(state, context):
     return dict(contents=dict((context.id(key), context.id(value))
-                for key, value in state.items()))
+                              for key, value in state.items()))
 
 
 @loader(dict)
@@ -593,12 +621,16 @@ def _load_subset_state(rec, context):
 
 @saver(RangeSubsetState)
 def _save_range_subset_state(state, context):
-    return dict(lo=state.lo, hi=state.hi, att=context.id(state.att))
+    return dict(lo=context.id(state.lo),
+                hi=context.id(state.hi),
+                att=context.id(state.att))
 
 
 @loader(RangeSubsetState)
 def _load_range_subset_state(rec, context):
-    return RangeSubsetState(rec['lo'], rec['hi'], context.object(rec['att']))
+    return RangeSubsetState(context.object(rec['lo']),
+                            context.object(rec['hi']),
+                            context.object(rec['att']))
 
 
 @saver(RoiSubsetState)
@@ -693,11 +725,66 @@ def _save_data_collection_3(dc, context):
     return result
 
 
+@saver(DataCollection, version=4)
+def _save_data_collection_4(dc, context):
+    cids = [c for data in dc for c in data.component_ids()]
+    components = [data.get_component(c)
+                  for data in dc for c in data.component_ids()]
+    return dict(data=list(map(context.id, dc)),
+                links=list(map(context.id, dc.external_links)),
+                cids=list(map(context.id, cids)),
+                components=list(map(context.id, components)),
+                groups=list(map(context.id, dc.subset_groups)),
+                subset_group_count=dc._sg_count)
+
+
 @loader(DataCollection)
 def _load_data_collection(rec, context):
-    dc = DataCollection(list(map(context.object, rec['data'])))
-    for link in rec['links']:
-        dc.add_link(context.object(link))
+
+    datasets = list(map(context.object, rec['data']))
+
+    links = [context.object(link) for link in rec['links']]
+
+    # Filter out CoordinateComponentLinks that may have been saved in the past
+    # as these are now re-generated on-the-fly.
+    links = [link for link in links if not isinstance(link, CoordinateComponentLink)]
+
+    # Go through and split links into links internal to datasets and ones
+    # between datasets as this dictates whether they should be set on the
+    # data collection or on the data objects.
+    external, internal = [], []
+    for link in links:
+        parent_to = link.get_to_id().parent
+        for cid in link.get_from_ids():
+            if cid.parent is not parent_to:
+                external.append(link)
+                break
+        else:
+            internal.append(link)
+
+    # Remove components in datasets that have external links
+    for data in datasets:
+        remove = []
+        for cid in data.derived_components:
+            comp = data.get_component(cid)
+
+            # Neihter in external nor in links overall
+            if rec.get('_protocol', 0) <= 3:
+                if comp.link not in internal:
+                    remove.append(cid)
+
+            if isinstance(comp.link, CoordinateComponentLink):
+                remove.append(cid)
+
+            if len(comp.link.get_from_ids()) == 1 and comp.link.get_from_ids()[0].parent is comp.link.get_to_id().parent and comp.link.get_from_ids()[0].label == comp.link.get_to_id().label:
+                remove.append(cid)
+
+        for cid in remove:
+            data.remove_component(cid)
+
+    dc = DataCollection(datasets)
+
+    dc.set_links(external)
     coerce_subset_groups(dc)
     return dc
 
@@ -710,6 +797,7 @@ def _load_data_collection_2(rec, context):
         grp.register_to_hub(result.hub)
     return result
 
+
 @loader(DataCollection, version=3)
 def _load_data_collection_3(rec, context):
     result = _load_data_collection_2(rec, context)
@@ -717,6 +805,23 @@ def _load_data_collection_3(rec, context):
     return result
 
 
+@loader(DataCollection, version=4)
+def _load_data_collection_4(rec, context):
+
+    dc = DataCollection(list(map(context.object, rec['data'])))
+    links = [context.object(link) for link in rec['links']]
+    dc.set_links(links)
+    coerce_subset_groups(dc)
+
+    dc._subset_groups = list(map(context.object, rec['groups']))
+    for grp in dc.subset_groups:
+        grp.register_to_hub(dc.hub)
+
+    dc._sg_count = rec['subset_group_count']
+
+    return dc
+
+
 @saver(Data)
 def _save_data(data, context):
 
@@ -743,7 +848,7 @@ def _load_data(rec, context):
 
     # we manually rebuild pixel/world components, so
     # we override this function. This is pretty ugly
-    result._create_pixel_and_world_components = lambda: None
+    result._create_pixel_and_world_components = lambda ndim: None
 
     comps = [list(map(context.object, [cid, comp]))
              for cid, comp in rec['components']]
@@ -757,8 +862,7 @@ def _load_data(rec, context):
             # and upgrade it to one. This can be removed once we no longer
             # support pre-v0.8 session files.
             if not comp.world and not isinstance(cid, PixelComponentID):
-                cid = PixelComponentID(comp.axis, cid.label,
-                                       hidden=cid.hidden, parent=cid.parent)
+                cid = PixelComponentID(comp.axis, cid.label, parent=cid.parent)
                 comps[icomp] = (cid, comp)
 
         result.add_component(comp, cid)
@@ -773,6 +877,9 @@ def _load_data(rec, context):
     result._world_component_ids = coord[:len(coord) // 2]
     result._pixel_component_ids = coord[len(coord) // 2:]
 
+    # We can now re-generate the coordinate links
+    result._set_up_coordinate_component_links(result.ndim)
+
     for s in rec['subsets']:
         result.add_subset(context.object(s))
 
@@ -806,8 +913,10 @@ def _load_data_3(rec, context):
 @saver(Data, version=4)
 def _save_data_4(data, context):
     result = _save_data_2(data, context)
+
     def save_cid_tuple(cids):
         return tuple(context.id(cid) for cid in cids)
+
     result['_key_joins'] = [[context.id(k), save_cid_tuple(v0), save_cid_tuple(v1)]
                             for k, (v0, v1) in data._key_joins.items()]
     result['uuid'] = data.uuid
@@ -818,8 +927,10 @@ def _save_data_4(data, context):
 def _load_data_4(rec, context):
     result = _load_data_2(rec, context)
     yield result
+
     def load_cid_tuple(cids):
         return tuple(context.object(cid) for cid in cids)
+
     result._key_joins = dict((context.object(k), (load_cid_tuple(v0), load_cid_tuple(v1)))
                              for k, v0, v1 in rec['_key_joins'])
     if 'uuid' in rec and rec['uuid'] is not None:
@@ -832,6 +943,7 @@ def _load_data_4(rec, context):
 def _save_data_5(data, context):
     result = _save_data_4(data, context)
     result['primary_owner'] = [context.id(cid) for cid in data.components if cid.parent is data]
+    result['meta'] = context.do(data.meta)
     return result
 
 
@@ -843,29 +955,33 @@ def _load_data_5(rec, context):
             cid = context.object(cid)
             cid.parent = result
     yield result
+
     def load_cid_tuple(cids):
         return tuple(context.object(cid) for cid in cids)
+
     result._key_joins = dict((context.object(k), (load_cid_tuple(v0), load_cid_tuple(v1)))
                              for k, v0, v1 in rec['_key_joins'])
     if 'uuid' in rec and rec['uuid'] is not None:
         result.uuid = rec['uuid']
     else:
         result.uuid = str(uuid.uuid4())
+    if 'meta' in rec:
+        result.meta.update(context.object(rec['meta']))
 
 
 @saver(ComponentID)
 def _save_component_id(cid, context):
-    return dict(label=cid.label, hidden=cid.hidden)
+    return dict(label=cid.label)
 
 
 @loader(ComponentID)
 def _load_component_id(rec, context):
-    return ComponentID(rec['label'], rec['hidden'])
+    return ComponentID(rec['label'])
 
 
 @saver(PixelComponentID)
 def _save_pixel_component_id(cid, context):
-    return dict(axis=cid.axis, label=cid.label, hidden=cid.hidden)
+    return dict(axis=cid.axis, label=cid.label)
 
 
 @loader(PixelComponentID)
@@ -874,7 +990,7 @@ def _load_pixel_component_id(rec, context):
         axis = rec['axis']
     else:  # backward-compatibility
         axis = int(rec['label'].split()[2])
-    return PixelComponentID(axis, rec['label'], rec['hidden'])
+    return PixelComponentID(axis, rec['label'])
 
 
 @saver(Component)
@@ -891,11 +1007,15 @@ def _save_component(component, context):
 
 @loader(Component)
 def _load_component(rec, context):
+
     if 'log' in rec:
         return context.object(rec['log']).component(rec['log_item'])
 
-    return Component(data=context.object(rec['data']),
-                     units=rec['units'])
+    cls = lookup_class_with_patches(rec['_type'])
+
+    return cls(data=context.object(rec['data']),
+               units=rec['units'])
+
 
 @saver(CategoricalComponent)
 def _save_categorical_component(component, context):
@@ -921,6 +1041,7 @@ def _load_categorical_component(rec, context):
                                 jitter=context.object(rec['jitter_method']),
                                 units=rec['units'])
 
+
 @saver(DerivedComponent)
 def _save_derived_component(component, context):
     return dict(link=context.id(component.link))
@@ -937,8 +1058,7 @@ def _save_component_link(link, context):
     to = list(map(context.id, [link.get_to_id()]))
     using = context.do(link.get_using())
     inverse = context.do(link.get_inverse())
-    hidden = link.hidden
-    return dict(frm=frm, to=to, using=using, inverse=inverse, hidden=hidden)
+    return dict(frm=frm, to=to, using=using, inverse=inverse)
 
 
 @loader(ComponentLink)
@@ -948,7 +1068,6 @@ def _load_component_link(rec, context):
     using = context.object(rec['using'])
     inverse = context.object(rec['inverse'])
     result = ComponentLink(frm, to, using, inverse)
-    result.hidden = rec['hidden']
     return result
 
 
@@ -973,6 +1092,17 @@ def _load_coordinate_component_link(rec, context):
     return CoordinateComponentLink(frm, to, coords, index, pix2world)
 
 
+@saver(types.BuiltinFunctionType)
+def _save_builtin_function(function, context):
+    ref = "%s.%s" % (function.__module__, function.__name__)
+    return {'function': ref}
+
+
+@loader(types.BuiltinFunctionType)
+def _load_builtin_function(rec, context):
+    return lookup_class_with_patches(rec['function'])
+
+
 @saver(types.FunctionType)
 def _save_function(function, context):
     ref = "%s.%s" % (function.__module__, function.__name__)
@@ -1020,10 +1150,22 @@ def _save_numpy(obj, context):
     data = b64encode(f.getvalue()).decode('ascii')
     return dict(data=data)
 
+
 @saver(Colormap)
 def _save_cmap(cmap, context):
-    return {'cmap':cmap.name}
+    return {'cmap': cmap.name}
+
 
 @loader(Colormap)
 def _load_cmap(rec, context):
     return cm.get_cmap(rec['cmap'])
+
+
+@saver(np.datetime64)
+def _save_datetime64(dt, context):
+    return {'datetime64': str(dt)}
+
+
+@loader(np.datetime64)
+def _load_datetime64(rec, context):
+    return np.datetime64(rec['datetime64'])
