@@ -2,10 +2,7 @@ from __future__ import absolute_import, division, print_function
 
 import numpy as np
 
-from glue.viewers.matplotlib.qt.toolbar import MatplotlibViewerToolbar
-from glue.core.edit_subset_mode import EditSubsetMode
 from glue.core.roi import PointROI
-from glue.core import command
 from glue.core.subset import CategorySubsetState
 from glue.core.exceptions import IncompatibleDataException
 from glue.utils.qt import messagebox_on_error
@@ -25,7 +22,6 @@ class DendrogramViewer(MatplotlibDataViewer):
 
     LABEL = 'Dendrogram'
 
-    _toolbar_cls = MatplotlibViewerToolbar
     _layer_style_widget_cls = DendrogramLayerStyleEditor
     _state_cls = DendrogramViewerState
     _options_cls = DendrogramOptionsWidget
